package treeMapping;

import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.PostfixExpression;
import org.eclipse.jdt.core.dom.PrefixExpression;

import util.Pair;
import util.PairDescendingOrder;

public class TreeMapper {
	private MapSourceFile fileM, fileN;
	public static final String propertyRevisionMap = "rMap";
	public static final String propertySimilarity = "similarity";
	public static double simThreshold = 0.65;
	private ASTNode astM, astN;
	private HashMap<ASTNode, ArrayList<ASTNode>> treeM, treeN;
	
	public TreeMapper(MapSourceFile fileM, MapSourceFile fileN)
	{
		this.fileM = fileM;
		this.fileN = fileN;
		/*astM = fileM.getAst();
		astN = fileN.getAst();
		treeM = new HashMap<ASTNode, ArrayList<ASTNode>>(fileM.getTree());
		treeN = new HashMap<ASTNode, ArrayList<ASTNode>>(fileN.getTree());*/
	}
	/**
	 * 
	 * @return [0]: average of #nodes, [1]: #relabels, [2]: #dels, [3]: #adds
	 */
	public int[] map()
	{
		mapLines();
		
		fileM.buildTree();
		fileN.buildTree();
		
		astM = fileM.getAst();
		astN = fileN.getAst();
		treeM = new HashMap<ASTNode, ArrayList<ASTNode>>(fileM.getTree());
		treeN = new HashMap<ASTNode, ArrayList<ASTNode>>(fileN.getTree());
		
		mapIsomorphicLeaves(fileM.getAllLeaves(), fileN.getAllLeaves());
		/*System.out.println(fileM.getAllLeaves());
		System.out.println(fileN.getAllLeaves());*/
		
		MapVectorVisitor vectorVisitor = new MapVectorVisitor();
		astM.accept(vectorVisitor);
		vectorVisitor = new MapVectorVisitor();
		astN.accept(vectorVisitor);
		vectorVisitor = null;
		
		astM.setProperty(propertyRevisionMap, astN);
		astM.setProperty(propertySimilarity, 1.0);
		astN.setProperty(propertyRevisionMap, astM);
		astN.setProperty(propertySimilarity, 1.0);
		
		mapDeletion(astM);
		
		mapAddition(astN);
		
		mapMoving(astM);
		markAddition(astN);
		
		mapRelabeling(astM, treeM);
		/*mapNameRelabeling(astM, treeM);
		mapOperatorRelabeling(astM, treeM);*/
		
		int[] count = new int[4];
		checkUnchange(astM, treeM);
		checkUnchange(astN, treeN);
		int[] countM = mapUnchange(astM, treeM);
		int[] countN = mapUnchange(astN, treeN);
		count[0] = (countM[0] + countN[0]) / 2;
		count[1] = countM[1];
		count[2] = countM[2];
		count[3] = countN[2];
		
		return count;
	}
	private void mapLines()
	{
		ArrayList<Integer> unchangedLinesM = fileM.getUnchangedLines();
		ArrayList<Integer> unchangedLinesN = fileN.getUnchangedLines();
		if(unchangedLinesM.size() != unchangedLinesN.size())
		{
			System.out.println("Something wrong: unchanged lines are not the same");
			return;
		}
		HashSet<Integer> commentLinesM = fileM.getCommentLines();
		HashSet<Integer> commentLinesN = fileN.getCommentLines();
		int i = 0;
		while(i < unchangedLinesM.size())
		{
			if(commentLinesM.contains(unchangedLinesM.get(i)) || commentLinesN.contains(unchangedLinesN.get(i)))
			{
				unchangedLinesM.remove(i);
				unchangedLinesN.remove(i);
			}
			else
				i++;
		}
	}
	/**
	 * 
	 */
	private void mapIsomorphicLeaves(ArrayList<ASTNode> leavesM, ArrayList<ASTNode> leavesN)
	{
		int m = -1, n = -1;
		ArrayList<ASTNode> deletedLeaves = new ArrayList<ASTNode>(), addedLeaves = new ArrayList<ASTNode>();
		while(m < leavesM.size() || n < leavesN.size())
		{
			int prevM = m, prevN = n;
			m++; n++;
			boolean hasNeighbor = false;
			ASTNode leafM = null;
			ASTNode leafN = null;
			while(!hasNeighbor && (m < leavesM.size() || n < leavesN.size()))
			{
				while(m < leavesM.size())
				{
					leafM = leavesM.get(m);
					int status = (Integer)leafM.getProperty(MapVisitor.propertyStatus);
					if(status == MapVisitor.STATUS_UNCHANGED)
						break;
					m++;
				}
				while(n < leavesN.size())
				{
					leafN = leavesN.get(n);
					int status = (Integer)leafN.getProperty(MapVisitor.propertyStatus);
					if(status == MapVisitor.STATUS_UNCHANGED)
						break;
					n++;
				}
				if(m < leavesM.size() && n < leavesN.size())
				{
					hasNeighbor = false;
					/*if(!leafM.toString().equals(leafN.toString()))
					{
						System.out.println(leafM.getProperty(MapVisitor.propertyLocationID) + "@" + 
						leafM.getProperty(MapVisitor.propertyStartLine));
						System.out.println(leafM);
						System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
						System.out.println(leafN.getProperty(MapVisitor.propertyLocationID) + "@" + 
								leafN.getProperty(MapVisitor.propertyStartLine));
						System.out.println(leafN);
					}*/
					if(m > 0 && (Integer)leavesM.get(m-1).getProperty(MapVisitor.propertyStatus) == MapVisitor.STATUS_UNCHANGED
							//|| m < leavesM.size()-1 && (Integer)leavesM.get(m+1).getProperty(MapVisitor.propertyStatus) == MapVisitor.STATUS_UNCHANGED
							)
						hasNeighbor = true;
					int nodeTypeM = leafM.getNodeType();
					if(nodeTypeM != ASTNode.RETURN_STATEMENT && nodeTypeM != ASTNode.BREAK_STATEMENT && nodeTypeM != ASTNode.CONTINUE_STATEMENT && 
							getParent(leafM).getNodeType() != ASTNode.RETURN_STATEMENT && 
							m < leavesM.size()-1 && 
							(Integer)leavesM.get(m+1).getProperty(MapVisitor.propertyStatus) == MapVisitor.STATUS_UNCHANGED)
						hasNeighbor = true;
					if(!hasNeighbor)
					{
						if(n > 0 && (Integer)leavesN.get(n-1).getProperty(MapVisitor.propertyStatus) == MapVisitor.STATUS_UNCHANGED)
							hasNeighbor = true;
						int nodeTypeN = leafN.getNodeType();
						if(nodeTypeN != ASTNode.RETURN_STATEMENT && nodeTypeN != ASTNode.BREAK_STATEMENT && nodeTypeN != ASTNode.CONTINUE_STATEMENT &&
								getParent(leafN).getNodeType() != ASTNode.RETURN_STATEMENT && 
								n < leavesN.size()-1 && 
								(Integer)leavesN.get(n+1).getProperty(MapVisitor.propertyStatus) == MapVisitor.STATUS_UNCHANGED)
							hasNeighbor = true;
					}
					if(!hasNeighbor)
					{
						leafM.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_FULLY_CHANGED);
						leafN.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_FULLY_CHANGED);
					}
				}
				else if(m < leavesM.size() || n < leavesN.size())
				{
					System.out.println("Something wrong with isomorphic leaves HERE");
					if(m < leavesM.size())
					{
						leafM = leavesM.get(m);
						String pidM = (String)leafM.getProperty(MapVisitor.propertyLocationID);
						System.out.println("\n\n");
						System.out.println(pidM + "@" + leafM.getProperty(MapVisitor.propertyStartLine));
						System.out.println(leafM);
						
					}
					else
					{
						leafN = leavesN.get(n);
						String pidN = (String)leafN.getProperty(MapVisitor.propertyLocationID);
						System.out.println("\n\n");
						System.out.println(pidN + "@" + leafN.getProperty(MapVisitor.propertyStartLine));
						System.out.println(leafN);
					}
					return;
				}
			}
			
			if(m < leavesM.size() && n < leavesN.size())
			{
				leafM.setProperty(propertyRevisionMap, leafN);
				leafM.setProperty(propertySimilarity, 1.0);
				leafN.setProperty(propertyRevisionMap, leafM);
				leafN.setProperty(propertySimilarity, 1.0);
				/*System.out.println(leafM.getProperty(MapVisitor.propertyLocationID) + "@" + 
						leafM.getProperty(MapVisitor.propertyStartLine));
				System.out.println(leafM);
				System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
				System.out.println(leafN.getProperty(MapVisitor.propertyLocationID) + "@" + 
						leafN.getProperty(MapVisitor.propertyStartLine));
				System.out.println(leafN);*/
			}
			else if(m < leavesM.size() || n < leavesN.size())
			{
				System.out.println("Something wrong with isomorphic leaves");
				if(m < leavesM.size())
				{
					leafM = leavesM.get(m);
					String pidM = (String)leafM.getProperty(MapVisitor.propertyLocationID);
					System.out.println("\n\n");
					System.out.println(pidM + "@" + leafM.getProperty(MapVisitor.propertyStartLine));
					System.out.println(leafM);
					
				}
				else
				{
					leafN = leavesN.get(n);
					String pidN = (String)leafN.getProperty(MapVisitor.propertyLocationID);
					System.out.println("\n\n");
					System.out.println(pidN + "@" + leafN.getProperty(MapVisitor.propertyStartLine));
					System.out.println(leafN);
				}
				return;
			}
			int lenM = (m-1) - (prevM+1) + 1;
			int lenN = (n-1) - (prevN+1) + 1;
			if(lenM > 0 && lenN > 0)
			{
				int[][] d = new int[lenM+1][lenN+1];
				String[] codeM = new String[lenM+1];
				String[] codeN = new String[lenN+1];
				String[][] p = new String[lenM+1][lenN+1];
				d[0][0] = 0;
				for(int i = 1; i <= lenM; i++)
				{
					d[i][0] = 0;
					codeM[i] = leavesM.get(i + prevM).toString();
					/*System.out.println(i + " - " + codeM[i] + "@" + 
							leavesM.get(i + prevM).getProperty(MapVisitor.propertyStartLine) + "@" +
							leavesM.get(i + prevM).getProperty(MapVisitor.propertyStatus));*/
				}
				for(int i = 1; i <= lenN; i++)
				{
					d[0][i] = 0;
					codeN[i] = leavesN.get(i + prevN).toString();
					/*System.out.println(i + " - " + codeN[i] + "@" + 
							leavesN.get(i + prevN).getProperty(MapVisitor.propertyStartLine) + "@" +
							leavesN.get(i + prevN).getProperty(MapVisitor.propertyStatus));*/
				}
				for(int i = 1; i <= lenM; i++)
				{
					for(int j = 1; j <= lenN; j++)
					{
						if(codeM[i].equals(codeN[j])
								&& ((i>1 && j>1 && codeM[i-1].equals(codeN[j-1])) || (i<lenM && j<lenN && codeM[i+1].equals(codeN[j+1])))
								)
						{
							d[i][j] = d[i-1][j-1] + 1;
							p[i][j] = "LU";
						}
						else if(d[i-1][j] >= d[i][j-1])
						{
							d[i][j] = d[i-1][j];
							p[i][j] = "U";
						}
						else
						{
							d[i][j] = d[i][j-1];
							p[i][j] = "L";
						}
					}
				}
				ArrayList<Integer> lcsM = new ArrayList<Integer>();
				ArrayList<Integer> lcsN = new ArrayList<Integer>();
				printLCS(p, lenM, lenN, lcsM, lcsN);
				int pM = prevM, pN = prevN;
				for(int i = 0; i < lcsM.size(); i++)
				{
					//System.out.println(lcsM.get(i) + " - " + lcsN.get(i));
					int cM = lcsM.get(i) + prevM, cN = lcsN.get(i) + prevN;
					leafM = leavesM.get(cM);
					leafN = leavesN.get(cN);
					leafM.setProperty(propertyRevisionMap, leafN);
					leafM.setProperty(propertySimilarity, 1.0);
					leafN.setProperty(propertyRevisionMap, leafM);
					leafN.setProperty(propertySimilarity, 1.0);
					/*System.out.println(leafM.getProperty(MapVisitor.propertyLocationID) + "@" + 
							leafM.getProperty(MapVisitor.propertyStartLine));
					System.out.println(leafM);
					System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
					System.out.println(leafN.getProperty(MapVisitor.propertyLocationID) + "@" + 
							leafN.getProperty(MapVisitor.propertyStartLine));
					System.out.println(leafN);*/
					if(cM - pM == 1 && cN - pN > 1)
						for(int j = pN+1; j < cN; j++)
						{
							leavesN.get(j).setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_ADDED);
							addedLeaves.add(leavesN.get(j));
							/*System.out.println(leavesN.get(j).getProperty(MapVisitor.propertyLocationID) + "@" + 
									leavesN.get(j).getProperty(MapVisitor.propertyStartLine) + " - Added in mapping iso leaves");
							System.out.println(leavesN.get(j));*/
						}
					if(cN - pN == 1 && cM - pM > 1)
						for(int j = pM+1; j < cM; j++)
						{
							leavesM.get(j).setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_DELETED);
							deletedLeaves.add(leavesM.get(j));
							/*System.out.println(leavesM.get(j).getProperty(MapVisitor.propertyLocationID) + "@" + 
									leavesM.get(j).getProperty(MapVisitor.propertyStartLine) + " - Deleted in mapping iso leaves");
							System.out.println(leavesM.get(j));*/
						}
					pM = cM; pN = cN;
				}
				if(m - pM == 1 && n - pN > 1)
					for(int j = pN+1; j < n; j++)
					{
						leavesN.get(j).setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_ADDED);
						addedLeaves.add(leavesN.get(j));
						/*System.out.println(leavesN.get(j).getProperty(MapVisitor.propertyLocationID) + "@" + 
								leavesN.get(j).getProperty(MapVisitor.propertyStartLine) + " - Added in mapping iso leaves");
						System.out.println(leavesN.get(j));*/
					}
				if(n - pN == 1 && m - pM > 1)
					for(int j = pM+1; j < m; j++)
					{
						leavesM.get(j).setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_DELETED);
						deletedLeaves.add(leavesM.get(j));
						/*System.out.println(leavesM.get(j).getProperty(MapVisitor.propertyLocationID) + "@" + 
								leavesM.get(j).getProperty(MapVisitor.propertyStartLine) + " - Deleted in mapping iso leaves");
						System.out.println(leavesM.get(j));*/
					}
			}
			else if(lenN > 0)
			{
				for(int j = prevN+1; j < n; j++)
				{
					leavesN.get(j).setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_ADDED);
					addedLeaves.add(leavesN.get(j));
					/*System.out.println(leavesN.get(j).getProperty(MapVisitor.propertyLocationID) + "@" + 
							leavesN.get(j).getProperty(MapVisitor.propertyStartLine) + " - Added in mapping iso leaves");
					System.out.println(leavesN.get(j));*/
				}
			}
			else if(lenM > 0)
			{
				for(int j = prevM+1; j < m; j++)
				{
					leavesM.get(j).setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_DELETED);
					deletedLeaves.add(leavesM.get(j));
					/*System.out.println(leavesM.get(j).getProperty(MapVisitor.propertyLocationID) + "@" + 
							leavesM.get(j).getProperty(MapVisitor.propertyStartLine) + " - Deleted in mapping iso leaves");
					System.out.println(leavesM.get(j));*/
				}
			}
		}
		
	}
	private void printLCS(String[][] p, int i, int j, ArrayList<Integer> lcsM, ArrayList<Integer> lcsN)
	{
		/*if(i == 0 || j == 0)
			return;
		if(p[i][j].equals("LU"))
		{
			printLCS(p, i-1, j-1, lcsM, lcsN);
			lcsM.add(i);
			lcsN.add(j);
		}
		else if(p[i][j].equals("U"))
			printLCS(p, i-1, j, lcsM, lcsN);
		else
			printLCS(p, i, j-1, lcsM, lcsN);*/

		int ii = i, jj = j;
		while(ii != 0 && jj != 0)
		{
			if(p[ii][jj].equals("LU"))
			{
				lcsM.add(0, ii);
				lcsN.add(0, jj);
				ii--; jj--;
			}
			else if(p[ii][jj].equals("U"))
				ii--;
			else
				jj--;
		}
	}
	/**
	 * 
	 */
	private void mapDeletion(ASTNode nodeM)
	{
		ArrayList<ASTNode> childrenM = treeM.get(nodeM);
		if(!childrenM.isEmpty())
		{
			for(int i = 0; i < childrenM.size(); i++)
			{
				ASTNode child = childrenM.get(i);
				mapDeletion(child);

			}
			HashSet<ASTNode> nodesN = new HashSet<ASTNode>();
			for(ASTNode childM : childrenM)
			{
				ASTNode childN = (ASTNode)childM.getProperty(propertyRevisionMap);
				if(childN != null)
					if(getParent(childN) != null)
						nodesN.add(getParent(childN));
			}
			if(nodesN.isEmpty())
			{
				nodeM.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_DELETED);
				/*System.out.println(nodeM.getProperty(MapVisitor.propertyLocationID) + "@" + 
						nodeM.getProperty(MapVisitor.propertyStartLine) + " - Deleted by no nodes");
				System.out.println(nodeM);*/
			}
			/*else if(nodesN.size() == 1)
			{
				ASTNode simNode = (ASTNode)nodesN.toArray()[0];
				nodeM.setProperty(propertyMap, simNode);
				simNode.setProperty(propertyMap, nodeM);
				System.out.println(nodeM.getProperty(MapVisitor.propertyLocationID) + "@" + 
						nodeM.getProperty(MapVisitor.propertyStartLine));
				System.out.println(nodeM);
				System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
				System.out.println(simNode.getProperty(MapVisitor.propertyLocationID) + "@" + 
						simNode.getProperty(MapVisitor.propertyStartLine));
				System.out.println(simNode);
			}*/
			else
			{
				MapFragment fragmentM = (MapFragment)nodeM.getProperty(MapVectorVisitor.propertyFragment);
				double similarity = 0;
				ASTNode simNode = null;
				for(ASTNode nodeN : nodesN)
				{
					ASTNode next = nodeN;
					do {
						MapFragment fragmentN = (MapFragment)next.getProperty(MapVectorVisitor.propertyFragment);
						double sim = fragmentM.changeSimilarity(fragmentN);
						//System.out.println(sim);
						if(next.getNodeType() == nodeM.getNodeType() && sim >= simThreshold && sim > similarity)
						//if(sim > similarity)
						{
							Double simM = (Double)nodeM.getProperty(propertySimilarity);
							Double simN = (Double)next.getProperty(propertySimilarity);
							if(simM == null)
								simM = 0.0;
							if(simN == null)
								simN = 0.0;
							if(simM < sim && simN < sim)
							{
								similarity = sim;
								simNode = next;
								if(sim == 1.0)
									break;
							}
						}
						next = getParent(next);
					} while(next != null && next.getNodeType() != ASTNode.METHOD_DECLARATION);
					if(similarity == 1.0)
						break;
				}
				if(simNode != null)
				{
					{
						ASTNode mappedNode = (ASTNode)nodeM.getProperty(propertyRevisionMap);
						if(mappedNode != null && (ASTNode)mappedNode.getProperty(propertyRevisionMap) == nodeM)
						{
							mappedNode.setProperty(propertyRevisionMap, null);
							mappedNode.setProperty(propertySimilarity, null);
						}
						nodeM.setProperty(propertyRevisionMap, simNode);
						nodeM.setProperty(propertySimilarity, similarity);
						mappedNode = (ASTNode)simNode.getProperty(propertyRevisionMap);
						if(mappedNode != null && (ASTNode)mappedNode.getProperty(propertyRevisionMap) == simNode)
						{
							mappedNode.setProperty(propertyRevisionMap, null);
							mappedNode.setProperty(propertySimilarity, null);
						}
						simNode.setProperty(propertyRevisionMap, nodeM);
						simNode.setProperty(propertySimilarity, similarity);
						/*System.out.println(nodeM.getProperty(MapVisitor.propertyLocationID) + "@" + 
								nodeM.getProperty(MapVisitor.propertyStartLine));
						System.out.println(nodeM);
						System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
						System.out.println(simNode.getProperty(MapVisitor.propertyLocationID) + "@" + 
								simNode.getProperty(MapVisitor.propertyStartLine));
						System.out.println(simNode);*/
					}
				}
				else
				{
					if(nodeM.getProperty(propertyRevisionMap) == null)
					{
						nodeM.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_DELETED);
						/*System.out.println(nodeM.getProperty(MapVisitor.propertyLocationID) + "@" + 
								nodeM.getProperty(MapVisitor.propertyStartLine) + " - Deleted by no similar nodes " + similarity);
						System.out.println(nodeM);*/
					}
				}
			}
		}
	}
	/**
	 * 
	 */
	private void mapAddition(ASTNode nodeN)
	{
		ArrayList<ASTNode> childrenN = treeN.get(nodeN);
		//System.out.println(nodeN);
		if(!childrenN.isEmpty())
		{
			for(int i = 0; i < childrenN.size(); i++)
			{
				ASTNode child = childrenN.get(i);
				mapAddition(child);

			}
			HashSet<ASTNode> nodesM = new HashSet<ASTNode>();
			for(ASTNode childN : childrenN)
			{
				ASTNode childM = (ASTNode)childN.getProperty(propertyRevisionMap);
				if(childM != null)
					if(getParent(childM) != null)
						nodesM.add(getParent(childM));
			}
			if(nodesM.isEmpty())
			{
				nodeN.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_ADDED);
				/*System.out.println(nodeN.getProperty(MapVisitor.propertyLocationID) + "@" + 
						nodeN.getProperty(MapVisitor.propertyStartLine) + " - Added");
				System.out.println(nodeN);*/
			}
			/*else if(nodesM.size() == 1)
			{
				ASTNode simNode = (ASTNode)nodesM.toArray()[0];
				nodeN.setProperty(propertyMap, simNode);
				simNode.setProperty(propertyMap, nodeN);
				System.out.println(nodeN.getProperty(MapVisitor.propertyLocationID) + "@" + 
						nodeN.getProperty(MapVisitor.propertyStartLine));
				System.out.println(nodeN);
				System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
				System.out.println(simNode.getProperty(MapVisitor.propertyLocationID) + "@" + 
						simNode.getProperty(MapVisitor.propertyStartLine));
				System.out.println(simNode);
			}*/
			else
			{
				MapFragment fragmentN = (MapFragment)nodeN.getProperty(MapVectorVisitor.propertyFragment);
				double similarity = 0;
				ASTNode simNode = null;
				for(ASTNode nodeM : nodesM)
				{
					ASTNode next = nodeM;
					do {
						MapFragment fragmentM = (MapFragment)next.getProperty(MapVectorVisitor.propertyFragment);
						double sim = fragmentN.changeSimilarity(fragmentM);
						if(next.getNodeType() == nodeN.getNodeType() && sim >= simThreshold && sim > similarity)
						//if(sim > similarity)
						{
							Double simM = (Double)next.getProperty(propertySimilarity);
							Double simN = (Double)nodeN.getProperty(propertySimilarity);
							if(simM == null)
								simM = 0.0;
							if(simN == null)
								simN = 0.0;
							if(simM < sim && simN < sim)
							{
								similarity = sim;
								simNode = next;
								if(sim == 1.0)
									break;
							}
						}
						next = getParent(next);
					} while(next != null && next.getNodeType() != ASTNode.METHOD_DECLARATION);
					if(similarity == 1.0)
						break;
				}
				if(simNode != null)
				{
					{
						ASTNode mappedNode = (ASTNode)nodeN.getProperty(propertyRevisionMap);
						if(mappedNode != null && (ASTNode)mappedNode.getProperty(propertyRevisionMap) == nodeN)
						{
							mappedNode.setProperty(propertyRevisionMap, null);
							mappedNode.setProperty(propertySimilarity, null);
						}
						nodeN.setProperty(propertyRevisionMap, simNode);
						nodeN.setProperty(propertySimilarity, similarity);
						mappedNode = (ASTNode)simNode.getProperty(propertyRevisionMap);
						if(mappedNode != null && (ASTNode)mappedNode.getProperty(propertyRevisionMap) == simNode)
						{
							mappedNode.setProperty(propertyRevisionMap, null);
							mappedNode.setProperty(propertySimilarity, null);
						}
						simNode.setProperty(propertyRevisionMap, nodeN);
						simNode.setProperty(propertySimilarity, similarity);
						/*System.out.println(nodeN.getProperty(MapVisitor.propertyLocationID) + "@" + 
								nodeN.getProperty(MapVisitor.propertyStartLine));
						System.out.println(nodeN);
						System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
						System.out.println(simNode.getProperty(MapVisitor.propertyLocationID) + "@" + 
								simNode.getProperty(MapVisitor.propertyStartLine));
						System.out.println(simNode);*/
					}
				}
				else
				{
					if(nodeN.getProperty(propertyRevisionMap) == null)
					{
						nodeN.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_ADDED);
						/*System.out.println(nodeN.getProperty(MapVisitor.propertyLocationID) + "@" + 
								nodeN.getProperty(MapVisitor.propertyStartLine) + " - Added");
						System.out.println(nodeN);*/
					}
				}
			}
		}
	}
	private void mapMoving(ASTNode rootM)
	{
		ArrayList<ASTNode> childrenM = treeM.get(rootM);
		ASTNode rootN = (ASTNode)rootM.getProperty(propertyRevisionMap);
		if(rootN != null)
		{
			ArrayList<ASTNode> childrenN = treeN.get(rootN);
			ArrayList<ASTNode> deletedChildren = new ArrayList<ASTNode>();
			ArrayList<ASTNode> addedChildren = new ArrayList<ASTNode>();
			for(ASTNode child : childrenM)
			{
				if(child.getProperty(propertyRevisionMap) == null)
					deletedChildren.add(child);
			}
			for(ASTNode child : childrenN)
			{
				if(child.getProperty(propertyRevisionMap) == null)
					addedChildren.add(child);
			}
			PairDescendingOrder comparator = new PairDescendingOrder();
			ArrayList<Pair> pairs = new ArrayList<Pair>();
			HashMap<ASTNode, HashSet<Pair>> pairsOf1 = new HashMap<ASTNode, HashSet<Pair>>(), pairsOf2 = new HashMap<ASTNode, HashSet<Pair>>();
			for(ASTNode nodeM : deletedChildren)
			{
				HashSet<Pair> pairs1 = pairsOf1.get(nodeM);
				if(pairs1 == null)
					pairs1 = new HashSet<Pair>();
				for(ASTNode nodeN : addedChildren)
				{
					Double[] sim = new Double[2];
					MapFragment fM = (MapFragment)nodeM.getProperty(MapVectorVisitor.propertyFragment);
					MapFragment fN = (MapFragment)nodeN.getProperty(MapVectorVisitor.propertyFragment);
					sim[0] = fM.nodeTypeSimilarity(fN);
					if(nodeM.getNodeType() == nodeN.getNodeType() && sim[0] >= simThreshold)
					{
						sim[1] = fM.leafLabelSimilarity(fN);
						int indexM = childrenM.indexOf(nodeM), indexN = childrenN.indexOf(nodeN);
						Pair pair = new Pair(nodeM, nodeN, sim[0]*1000000 + sim[1] + 0.0000001/(Math.abs(indexM-indexN)+1));
						int index = Collections.binarySearch(pairs, pair, comparator);
						if(index < 0)
							pairs.add(-1-index, pair);
						else
							pairs.add(index, pair);
						pairs1.add(pair);
						HashSet<Pair> pairs2 = pairsOf2.get(nodeN);
						if(pairs2 == null)
							pairs2 = new HashSet<Pair>();
						pairs2.add(pair);
						pairsOf2.put(nodeN, pairs2);
					}
				}
				pairsOf1.put(nodeM, pairs1);
			}
			while(!pairs.isEmpty())
			{
				Pair pair = pairs.get(0);
				ASTNode node1 = (ASTNode)pair.getObj1(), node2 = (ASTNode)pair.getObj2();
				node1.setProperty(propertyRevisionMap, node2);
				node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_MOVED);
				node2.setProperty(propertyRevisionMap, node1);
				node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_MOVED);
				for(Pair p : pairsOf1.get(node1))
					pairs.remove(p);
				for(Pair p : pairsOf2.get(node2))
					pairs.remove(p);
			}
			/*int lenM = childrenM.size(), lenN = childrenN.size();
			int[][] d = new int[lenM+1][lenN+1];
			ASTNode[] childM = new ASTNode[lenM+1];
			ASTNode[] childN = new ASTNode[lenN+1];
			String[][] p = new String[lenM+1][lenN+1];
			d[0][0] = 0;
			for(int i = 1; i <= lenM; i++)
			{
				d[i][0] = 0;
				childM[i] = childrenM.get(i-1);
			}
			for(int i = 1; i <= lenN; i++)
			{
				d[0][i] = 0;
				childN[i] = childrenN.get(i-1);
			}
			for(int i = 1; i <= lenM; i++)
			{
				for(int j = 1; j <= lenN; j++)
				{
					if(childM[i].equals(childN[j].getProperty(propertyRevisionMap))
							//&& ((i>1 && j>1 && childM[i-1].equals(childN[j-1].getProperty(propertyRevisionMap))) || (i<lenM && j<lenN && childM[i+1].equals(childN[j+1].getProperty(propertyRevisionMap))))
							)
					{
						d[i][j] = d[i-1][j-1] + 1;
						p[i][j] = "LU";
					}
					else if(d[i-1][j] >= d[i][j-1])
					{
						d[i][j] = d[i-1][j];
						p[i][j] = "U";
					}
					else
					{
						d[i][j] = d[i][j-1];
						p[i][j] = "L";
					}
				}
			}
			ArrayList<Integer> lcsM = new ArrayList<Integer>();
			ArrayList<Integer> lcsN = new ArrayList<Integer>();
			printLCS(p, lenM, lenN, lcsM, lcsN);
//			for(int i = 0; i < lcsM.size(); i++)
//			{
//				childM[lcsM.get(i)].setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_PARTLY_CHANGED);
//				childN[lcsN.get(i)].setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_PARTLY_CHANGED);
//			}
			for(int i = 1; i <= lenM; i++)
			{
				if(lcsM.contains(i))
					childM[i].setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_PARTLY_CHANGED);
				else
				{
					ASTNode mappedNode = (ASTNode)childM[i].getProperty(TreeMapper.propertyRevisionMap);
					if(mappedNode != null && childrenN.contains(mappedNode))
						childM[i].setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_MOVED);
				}
			}
			for(int j = 1; j <= lenN; j++)
			{
				if(lcsN.contains(j))
					childN[j].setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_PARTLY_CHANGED);
				else 
				{
					ASTNode mappedNode = (ASTNode)childN[j].getProperty(TreeMapper.propertyRevisionMap);
					if(mappedNode != null && childrenM.contains(mappedNode))
						childN[j].setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_MOVED);
				}
			}*/
		}
		else
		{
			rootM.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_DELETED);
		}
		for(ASTNode childM : childrenM)
			mapMoving(childM);
	}
	private void markAddition(ASTNode rootN)
	{
		if(rootN.getProperty(propertyRevisionMap) == null)
			rootN.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_ADDED);
		ArrayList<ASTNode> children = treeN.get(rootN);
		for(ASTNode child : children)
		{
			markAddition(child);
		}
	}
	/*private void mapNameRelabeling(ASTNode node1, HashMap<ASTNode, ArrayList<ASTNode>> tree1)
	{
		ArrayList<ASTNode> children = tree1.get(node1);
		if(children.isEmpty())
		{
			ASTNode node2 = (ASTNode)node1.getProperty(propertyRevisionMap);
			if(node2 != null)
				if(!node1.toString().equals(node2.toString()))
				{
					node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
					node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
				}
		}
		else
			for(ASTNode child : children)
				mapNameRelabeling(child, tree1);
	}*/
	/*private void mapNameRelabeling()
	{
		for(ASTNode leafM : fileM.getAllLeaves())
		{
			ASTNode leafN = (ASTNode)leafM.getProperty(propertyRevisionMap);
			if(leafN != null)
				if(!leafM.toString().equals(leafN.toString()))
				{
					leafM.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
					leafN.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
				}
		}
	}*/
	private void mapRelabeling(ASTNode node1, HashMap<ASTNode, ArrayList<ASTNode>> tree1)
	{
		ArrayList<ASTNode> children = tree1.get(node1);
		if(children.isEmpty())
		{
			ASTNode node2 = (ASTNode)node1.getProperty(propertyRevisionMap);
			if(node2 != null)
			{
				if(!node1.toString().equals(node2.toString()))
				{
					node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
					node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
				}
				else if((Integer)node1.getProperty(MapVisitor.propertyStatus) != MapVisitor.STATUS_MOVED)
				{
					node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
					node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
				}
			}
		}
		else
		{
			int nodeType = node1.getNodeType();
			if((nodeType == ASTNode.INFIX_EXPRESSION || nodeType == ASTNode.PREFIX_EXPRESSION || nodeType == ASTNode.POSTFIX_EXPRESSION) && 
					node1.getProperty(propertyRevisionMap) != null)
			{
				ASTNode node2 = (ASTNode)node1.getProperty(propertyRevisionMap);
				if(node2.getNodeType() == nodeType)
				{
					if(nodeType == ASTNode.INFIX_EXPRESSION)
					{
						if(!((InfixExpression)node1).getOperator().equals(((InfixExpression)node2).getOperator()))
						{
							node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
							node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
						}
						else
						{
							node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
							node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
						}
					}
					else if(nodeType == ASTNode.PREFIX_EXPRESSION)
					{
						if(!((PrefixExpression)node1).getOperator().equals(((PrefixExpression)node2).getOperator()))
						{
							node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
							node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
						}
						else
						{
							node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
							node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
						}
					}
					else if(nodeType == ASTNode.POSTFIX_EXPRESSION)
					{
						if(!((PostfixExpression)node1).getOperator().equals(((PostfixExpression)node2).getOperator()))
						{
							node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
							node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
						}
						else
						{
							node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
							node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
						}
					}
				}
			}
			for(ASTNode child : children)
				mapRelabeling(child, tree1);
		}
	}
	private boolean checkUnchange(ASTNode node, HashMap<ASTNode, ArrayList<ASTNode>> tree)
	{
		int status = (Integer)node.getProperty(MapVisitor.propertyStatus);
		boolean unchanged = status <= MapVisitor.STATUS_FULLY_CHANGED;
		ArrayList<ASTNode> children = tree.get(node);
		for(ASTNode child : children)
		{
			boolean unChild = checkUnchange(child, tree);
			unchanged = (unchanged && unChild);
		}
		if(unchanged)
			node.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_UNCHANGED);
		return unchanged;
	}
	/**
	 * 
	 * @param node
	 * @param tree
	 * @return [0]: number of nodes, [1]: #relabels, [2]: #dels or adds
	 */
	private int[] mapUnchange(ASTNode node, HashMap<ASTNode, ArrayList<ASTNode>> tree)
	{
		int[] count = {0, 0, 0};
		count[0]++;
		//System.err.println(node.getClass().getSimpleName());
		ASTNode mappedNode = (ASTNode)node.getProperty(propertyRevisionMap);
		int status = (Integer)node.getProperty(MapVisitor.propertyStatus);
		switch (status) {
		case MapVisitor.STATUS_RELABELED:
			count[1]++;
			break;
		case MapVisitor.STATUS_DELETED:
			count[2]++;
			break;
		case MapVisitor.STATUS_ADDED:
			count[2]++;
			break;
		}
		if(mappedNode != null)
		{
			int mappedStatus = (Integer)mappedNode.getProperty(MapVisitor.propertyStatus);
			/*if(node.getNodeType() == ASTNode.METHOD_DECLARATION)
			{
				MethodDeclaration method = (MethodDeclaration)node;
				if(method.getName().getIdentifier().equals("attachPosixApplicationInput"))
					System.err.println("Start checking attachPosixApplicationInput in tree mapping: " + status + " vs. " + mappedStatus);
			}*/
			if(status == MapVisitor.STATUS_UNCHANGED && mappedStatus != MapVisitor.STATUS_UNCHANGED)
				node.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_PARTLY_CHANGED);
		}
		ArrayList<ASTNode> children = tree.get(node);
		for(ASTNode child : children)
		{
			int[] sub = mapUnchange(child, tree);
			for(int i = 0; i < count.length; i++)
				count[i] += sub[i];
		}
		return count;
	}
	/*private void mapOperatorRelabeling(ASTNode node1, HashMap<ASTNode, ArrayList<ASTNode>> tree1)
	{
		int nodeType = node1.getNodeType();
		if((nodeType == ASTNode.INFIX_EXPRESSION || nodeType == ASTNode.PREFIX_EXPRESSION || nodeType == ASTNode.POSTFIX_EXPRESSION) && 
				node1.getProperty(propertyRevisionMap) != null)
		{
			ASTNode node2 = (ASTNode)node1.getProperty(propertyRevisionMap);
			if(node2.getNodeType() == nodeType)
			{
				if(nodeType == ASTNode.INFIX_EXPRESSION && !((InfixExpression)node1).getOperator().equals(((InfixExpression)node2).getOperator()))
				{
					node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
					node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
				}
				else if(nodeType == ASTNode.PREFIX_EXPRESSION && !((PrefixExpression)node1).getOperator().equals(((PrefixExpression)node2).getOperator()))
				{
					node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
					node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
				}
				else if(nodeType == ASTNode.POSTFIX_EXPRESSION && !((PostfixExpression)node1).getOperator().equals(((PostfixExpression)node2).getOperator()))
				{
					node1.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
					node2.setProperty(MapVisitor.propertyStatus, MapVisitor.STATUS_RELABELED);
				}
			}
		}
	}*/
	/**
	 * 
	 */
	private ASTNode getParent(ASTNode node)
	{
		ASTNode pNode = node.getParent();
		while(pNode != null && 
				(pNode.getNodeType() == ASTNode.BLOCK || pNode.getNodeType() == ASTNode.EXPRESSION_STATEMENT || pNode.getNodeType() == ASTNode.PARENTHESIZED_EXPRESSION))
			pNode = pNode.getParent();
		
		return pNode;
	}
	public void printTree(HashMap<ASTNode, ArrayList<ASTNode>> tree, ASTNode root)
	{
		int status = (Integer)root.getProperty(MapVisitor.propertyStatus);
		//if(status > MapVisitor.STATUS_FULLY_CHANGED)
		if(status > MapVisitor.STATUS_UNCHANGED)
		{
			ASTNode simNode = (ASTNode)root.getProperty(propertyRevisionMap);
			System.out.println(root.getProperty(MapVisitor.propertyLocationID) + "@" + 
					root.getProperty(MapVisitor.propertyStartLine) + ": " + getStatus(status));
			System.out.println(root);
			if(simNode != null)
			{
				System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
				System.out.println(simNode.getProperty(MapVisitor.propertyLocationID) + "@" + 
						simNode.getProperty(MapVisitor.propertyStartLine));
				System.out.println(simNode);
			}
		}
		ArrayList<ASTNode> children = tree.get(root);
		if(children == null)
			System.err.println(root);
		for(ASTNode child : children)
		{
			printTree(tree, child);
		}
	}
	private String getStatus(int status)
	{
		switch(status) {
		case MapVisitor.STATUS_ADDED: return "ADDED";
		case MapVisitor.STATUS_DELETED: return "DELETED";
		case MapVisitor.STATUS_FULLY_CHANGED: return "FULLY_CHANGED";
		case MapVisitor.STATUS_MOVED: return "MOVED";
		case MapVisitor.STATUS_PARTLY_CHANGED: return "PARTLY_CHANGED";
		case MapVisitor.STATUS_RELABELED: return "RELABELED";
		case MapVisitor.STATUS_UNCHANGED: return "UNCHANGED";
		}
		return "NOTHING_ELSE";
	}
}
